local MODNAME, ShuckIt = ...

-- ShuckIt.lua
-- $Id: ShuckIt.lua 75 2015-01-10 05:19:10Z LaoTseu $

-- The ShuckIt idea came from the Shucks module from Cork (http://www.wowinterface.com/downloads/info10629) by Tekkub.
-- The basic framework for this addon is based on Tekkub's AddonTemplate (http://github.com/tekkub/addontemplate/tree/master)
-- The code to find the openable items is based on Ammo's work in
-- CrowBar (http://www.wowace.com/addons/crow-bar/) and is used with his permission.


---------------------------
-- Localized Lua globals --
---------------------------

local _G = getfenv(0)

local assert = _G.assert
local error = _G.error
local floor = _G.floor
local geterrorhandler = _G.geterrorhandler
local ipairs = _G.ipairs
local pairs = _G.pairs
local rawset = _G.rawset
local select = _G.select
local setmetatable = _G.setmetatable
local sort = _G.sort
local strjoin = _G.strjoin
local tinsert = _G.tinsert
local tonumber = _G.tonumber
local tostring = _G.tostring
local tremove = _G.tremove
local wipe = _G.wipe

local CreateFrame = _G.CreateFrame
local GetContainerItemInfo = _G.GetContainerItemInfo
local GetContainerItemLink = _G.GetContainerItemLink
local GetContainerItemQuestInfo = _G.GetContainerItemQuestInfo
local GetContainerNumSlots = _G.GetContainerNumSlots
local GetCurrentMapAreaID = _G.GetCurrentMapAreaID
local GetItemCount = _G.GetItemCount
local GetItemInfo = _G.GetItemInfo
local GetSpellInfo = _G.GetSpellInfo
local GetRealZoneText = _G.GetRealZoneText
local InCombatLockdown = _G.InCombatLockdown
local IsQuestFlaggedCompleted = _G.IsQuestFlaggedCompleted
local IsStealthed = _G.IsStealthed
local UIParent = _G.UIParent
local UnitLevel = _G.UnitLevel
local UseContainerItem = _G.UseContainerItem
local WorldFrame = _G.WorldFrame

local ATTRIBUTE_NOOP = _G.ATTRIBUTE_NOOP
local ITEM_OPENABLE = _G.ITEM_OPENABLE
local LOCKED = _G.LOCKED
local NUM_BAG_FRAMES = _G.NUM_BAG_FRAMES

----------------------
--     Revision     --
----------------------


ShuckIt = ShuckIt or {}
ShuckIt.revision = ShuckIt.revision or {}
ShuckIt.revision.main	= ("$Revision: 75 $"):match("(%d+)")
ShuckIt.version      	= _G.GetAddOnMetadata("ShuckIt", "Version"):match("([^ ]+)")
ShuckIt.revision.toc  	= _G.GetAddOnMetadata("ShuckIt", "Version"):match("%$Revision:%s(%d+)")

----------------------
--      Locals      --
----------------------

local L = ShuckIt.localized
--local defaults, defaultsPC, db, dbpc = {}, {}
local UPDATEPERIOD, elapsed = 0.5, 0
local TOOLTIPPERIOD, tooltip_elapsed = 0.2, 0
local need_scan, need_chat_update, busy, delay_scan = nil, nil, nil, nil

local item_to_shuck, icon, shuck_count = nil, nil, 0
-- I'm creating three tables because I don't want to deal with recycling sub-table
-- The ideal data type would have been a table of triplets { itemID, bag, slot }
local shuck_table = {}
local shuck_table_bag = {}
local shuck_table_slot = {}
local shuck_bag, shuck_slot = 1, 1
local default_icon = "Interface\\Icons\\INV_Misc_Shell_03"
local found_quest_item
local HEROIC_OPEN = L["Use: Open the Heroic Cache of Treasures to earn your glorious reward! (1 Sec Cooldown)"]

-- Frames
local f
local secureframe

local lqt = _G.LibStub("LibQTip-1.0")
local ldb = _G.LibStub:GetLibrary("LibDataBroker-1.1")
local dataobj = ldb:NewDataObject("ShuckIt",
											 {
												type = "data source",
											  	text = "",
											  	icon = default_icon
})

------------------------------
--      Util Functions      --
------------------------------

local function err(msg,...) _G.geterrorhandler()(msg:format(_G.tostringall(...)) .. " - " .. _G.time()) end

local default_channel = nil
local function setDefaultChanelForPrint()
	default_channel = nil
	for i=1, _G.NUM_CHAT_WINDOWS do
		local name = _G.GetChatWindowInfo(i)
		if name and name:lower() == 'output' then
			default_channel = i
		end
	end
end

local function Print(message, ...)
	if default_channel then
		_G["ChatFrame"..default_channel]:AddMessage(("|cff00dbba"..L["ShuckIt"].."|r: "..message):format(...));
	else
		_G.SELECTED_CHAT_FRAME:AddMessage(("|cff00dbba"..L["ShuckIt"].."|r: "..message):format(...))
	end
end

local debugf = _G.tekDebug and _G.tekDebug:GetFrame("ShuckIt")
local function Debug(...) if debugf then debugf:AddMessage(strjoin(", ", ...)) end end

------------------------------
--      Core Functions     --
------------------------------

local ShuckItTooltip = _G.CreateFrame("GameTooltip", "ShuckItTooltip")
ShuckItTooltip:SetOwner(_G.WorldFrame, "ANCHOR_NONE")

ShuckItTooltip:AddFontStrings(
    ShuckItTooltip:CreateFontString( "$parentTextLeft1", nil, "GameTooltipText" ),
    ShuckItTooltip:CreateFontString( "$parentTextRight1", nil, "GameTooltipText" )
)

local knownOpenable = {
}
local cache = {}
local do_not_cache = {}
local can_be_combine = {}
local function IsOpenable(bag, slot, itemLink)


	if not bag or not slot then return false end
	if not itemLink then itemLink = GetContainerItemLink(bag, slot) end
	if not itemLink then return false end -- Bag slot empty

	-- Is this a quest item for a quest that has not been activated yet ?
	local _, questID, isActive = GetContainerItemQuestInfo(bag, slot)
	if cache[itemLink] ~= nil then return cache[itemLink] end

	-- Check for known openable items
	local itemID = itemLink:match("[|]Hitem:([^|:]*)")
	if knownOpenable[itemID] then return true end

	if questID then
		local minLevel = select(5, GetItemInfo(itemLink))
		if minLevel <= UnitLevel("player") then
			found_quest_item = true
			-- No caching since a quest can be abandoned
			if not isActive and not IsQuestFlaggedCompleted(questID) then
				return true
			else
				return false
			end
		end
	end

	-- Tooltip parsing
	ShuckItTooltip:ClearLines()
	if not ShuckItTooltip:IsOwned(WorldFrame) then ShuckItTooltip:SetOwner(WorldFrame, "ANCHOR_NONE") end
 	ShuckItTooltip:SetBagItem(bag,slot)

	for i=1,ShuckItTooltip:NumLines() do
	   local mytext = _G["ShuckItTooltipTextLeft" .. i]
	   local text = mytext:GetText()

		if (text == ITEM_OPENABLE) then
			if not do_not_cache[itemLink] then cache[itemLink] = true end
			return true
		elseif text == LOCKED then
			-- Do not cache the result for locked items, they may become unlocked
			do_not_cache[itemLink] = true
			return false
		end
	end

	-- Is the item lootable according to GetContainerItemInfo?
	-- This needs to be done after the tooltip parsing otherwise the locked items are not detected.
	--  F98D00 (unlocked) 6B5B900 (locked)
	local lootable = select(6, GetContainerItemInfo(bag, slot))
	if lootable then
		if not do_not_cache[itemLink] then cache[itemLink] = true end
		return true
	end

	if not do_not_cache[itemLink] then cache[itemLink] = false end
	return false

end

-- _G.IsOpenable = IsOpenable

-- Cache some items that cannot be detected. If anyone has a clue how to detect that these are openable, please let me know
-- These work like clams in the sence that UseContainerItem will not work on them
local knownClams = {
	["58856"]	= true,		-- Royal Monkfish
	["89125"]	= true,		-- Sack of pet supplies
	["91086"]	= true,		-- Darkmoon Pet Supplies
	["93146"]	= true,		-- Pandaren Spirit Pet Supplies
	["93147"]	= true,		-- Pandaren Spirit Pet Supplies
	["93148"]	= true,		-- Pandaren Spirit Pet Supplies
	["93149"]	= true,		-- Pandaren Spirit Pet Supplies
	["94207"]	= true,		-- Fabled Pandaren Pet Supplies
	["98095"]	= true,		-- Brawler's Pet
	["115981"]	= true, 		-- Abrogator Stone Cluster
	["120301"]	= true,		-- Armor Enchancement Token
	["120302"]	= true,		-- Weapon Enchancement Token
}

local isSalvageItem = {
	["114116"] = true,
	["114119"] = true,
	["114120"] = true,
}

local cache_clam = {}
local function IsAClam(bag, slot, itemLink)
	if not bag or not slot then return false end
	if not itemLink then itemLink = _G.GetContainerItemLink(bag, slot) end
	if not itemLink then return false end -- Bag slot empty

	-- Items that we cannot detect but that are openable
	local itemID = itemLink:match("[|]Hitem:([^|:]*)")
	if knownClams[itemID] then return true end

	if cache_clam[itemLink] ~= nil then return cache_clam[itemLink] end

	-- Is this a garrison salvage item?
	if isSalvageItem[itemID] then
--		if GetRealZoneText() == L["Salvage Yard"] then
		if GetCurrentMapAreaID() == 971 or GetCurrentMapAreaID() == 976 then
			return true -- No cache
		else
			return false -- No cache
		end
	end
	-- local p
	-- if itemID == "114057" or (bag == 0 and slot == 1) then
	--  	err("bag = %s, slot = %s, itemLink = %s",bag, slot, itemLink)
	--  	p = true
	-- end
	ShuckItTooltip:ClearLines()
	if not ShuckItTooltip:IsOwned(WorldFrame) then ShuckItTooltip:SetOwner(WorldFrame, "ANCHOR_NONE") end
 	ShuckItTooltip:SetBagItem(bag,slot)

 	-- if p then err("ShuckItTooltip:NumLines() = %s", ShuckItTooltip:NumLines()) end
	for i=1,ShuckItTooltip:NumLines() do
	   local mytext = _G["ShuckItTooltipTextLeft" .. i]
	   local text = mytext:GetText()
	   -- if p then err("text = %s", text) end

	   if text == L["Use: Open the clam!"] or text == HEROIC_OPEN or text:find(L["Use: Create a soulbound item appropriate for your loot specialization"]) then
	   	-- Validate that the legel requirement is met
	   	local reqLevel = select(5, GetItemInfo(itemLink))
	   	if reqLevel <= UnitLevel("player") then
				cache_clam[itemLink] = true
				return true
			end
		elseif text:find(L["Use: Combine ten"]) then
			do_not_cache[itemLink] = true
			can_be_combine[itemID] = true
			if GetItemCount(itemLink) >= 10 then
				return true
			end
		end
	end

	if not do_not_cache[itemLink] then cache_clam[itemLink] = false end
	return false
end
-- EBIsAClam = IsAClam

local function ScanBags()

	local item_to_shuck_id
	item_to_shuck_id, shuck_bag, shuck_slot = nil, nil, nil
	shuck_count = 0
	found_quest_item = nil
	wipe(shuck_table)
	wipe(shuck_table_bag)
	wipe(shuck_table_slot)

	local newIcon = nil
	for bag = 0, NUM_BAG_FRAMES, 1 do
		for slot = 1, GetContainerNumSlots(bag), 1 do

			local itemLink = GetContainerItemLink(bag, slot)
			if itemLink and (IsOpenable(bag, slot, itemLink) or IsAClam(bag, slot, itemLink)) then
				local itemID = itemLink:match("item:(%d+)")
				if not itemID then
					err("invalid item link: %s",itemLink)
				else
					shuck_table[itemID] = (shuck_table[itemID] or 0) + tonumber(select(2,GetContainerItemInfo(bag, slot)) or 1)
					-- we keep the bag and slot for the first item we find
					shuck_table_bag[itemID] = shuck_table_bag[itemID] or bag
					shuck_table_slot[itemID] = shuck_table_slot[itemID] or slot
					if not item_to_shuck_id or isSalvageItem[item_to_shuck_id] then
						item_to_shuck_id = itemID
						item_to_shuck = GetItemInfo(itemLink)
						shuck_bag, shuck_slot = bag, slot
						newIcon = GetContainerItemInfo(bag, slot)
					end
				end
			elseif itemLink and not GetItemInfo(itemLink) then
				-- This is a workaround for a intermitent situation when opening
				-- multiple items of the same kind in a row sometimes result in
				-- the item information being screewed up.
				-- If this happen, we delay de scan until the information is realiable
				delay_scan = true
				return
			end

		end
	end

	-- If there is a quest item, we must listen to quest log events
	if found_quest_item then
--err("found_quest_item = %s", found_quest_item)	
		f:RegisterEvent("UNIT_QUEST_LOG_CHANGED")
	else
		f:UnregisterEvent("UNIT_QUEST_LOG_CHANGED")
	end

	-- Round down the number of combinable items to 10
	for itemID, count in pairs(shuck_table) do
		if can_be_combine[itemID] then
			shuck_table[itemID] = floor(count/10)
		end
	end

	for k in pairs (shuck_table) do
		shuck_count = shuck_count + shuck_table[k]
	end
	icon = newIcon or default_icon
	need_scan = nil
end

local function myOnClick(self, button, down)
	if not InCombatLockdown() and not IsStealthed() and not busy and IsOpenable(shuck_bag, shuck_slot) then
		UseContainerItem(shuck_bag, shuck_slot)
		return true
	elseif not InCombatLockdown() and not IsStealthed() and not busy and IsAClam(shuck_bag, shuck_slot) then
		Print(L["Cannot open Clams by clicking on the LDB display. Use %s instead."],"|cffffd700/click ShuckItFrame|r")
	else
		return false
	end
end

-----------------------------
--      Event Handler      --
-----------------------------

f = _G.CreateFrame("frame")
f:SetScript("OnEvent", function(self, event, ...) if self[event] then return self[event](self, event, ...) end end)
f:RegisterEvent("ADDON_LOADED")


function f:ADDON_LOADED(event, addon)
	if addon ~= "ShuckIt" then return end

	self:UnregisterEvent("ADDON_LOADED")
	self.ADDON_LOADED = nil

	if _G.IsLoggedIn() then self:PLAYER_LOGIN() else self:RegisterEvent("PLAYER_LOGIN") end
end


function f:PLAYER_LOGIN()
	-- Create the secure frame
	ShuckIt:InitSecureFrame()

	-- Register events
	self:RegisterEvent("BAG_UPDATE")
	self:RegisterEvent("UNIT_SPELLCAST_SENT")
	self:RegisterEvent("UNIT_SPELLCAST_FAILED")
	self:RegisterEvent("UNIT_SPELLCAST_STOP")
	self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:RegisterEvent("UPDATE_CHAT_WINDOWS")
	self:RegisterEvent("PLAYER_LEVEL_UP")
	self:RegisterEvent("ZONE_CHANGED")
	self:RegisterEvent("ZONE_CHANGED_INDOORS")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA")

	-- Start the OnUpdate
	self:SetScript("OnUpdate", function(self, elap)
		elapsed = elapsed + elap
		tooltip_elapsed = tooltip_elapsed + elap

		-- Tooltip management
		if tooltip_elapsed >= TOOLTIPPERIOD then
			tooltip_elapsed = 0
			if not (ShuckIt.tooltip and ShuckIt.tooltip:IsMouseOver()) and
				not (ShuckIt.tooltip_anchor and ShuckIt.tooltip_anchor:IsMouseOver()) then
				ShuckIt.tooltip = lqt:Release(ShuckIt.tooltip)
				ShuckIt.tooltip_anchor = nil
			end
		end

		if delay_scan then
			-- Workaround for when good item information is not available and we
			-- need to wait a bit
			elapsed = UPDATEPERIOD/2
			need_scan = true
			delay_scan = nil
			return
		end

		if not (need_scan or need_chat_update) or busy or elapsed < UPDATEPERIOD or _G.InCombatLockdown() then return end


		elapsed = 0

		-- Print window if needed
		if need_chat_update then
			setDefaultChanelForPrint()
			need_chat_update = nil
		end

		-- Find object to Shuck
		ScanBags()

		-- Adjust the display
		if shuck_count > 0 then
			dataobj.text = (L["%s - %d"]):format(item_to_shuck, shuck_count)
		else
			dataobj.text = ""
		end
		dataobj.icon = icon

		-- Update the Tooltip
		if ShuckIt.tooltip then ShuckIt:DrawTooltip() end
	end)

	-- Set the OnClick function for the data broker
	dataobj.OnClick = myOnClick

	-- Find the revision number
	if ShuckIt.revision then
		ShuckIt.rev = 0
		for _, rev in pairs(ShuckIt.revision) do
			local revision = tonumber(rev)
			if revision and revision > ShuckIt.rev then ShuckIt.rev = revision end
		end
	else
		assert(false,"No ShuckIt.revision")
	end

	-- Find the default channel
	setDefaultChanelForPrint()

	need_scan = true

	self:UnregisterEvent("PLAYER_LOGIN")
	self.PLAYER_LOGIN = nil
end


function f:BAG_UPDATE(message, bagID)
--err("BAG_UPDATE")
	elapsed = 0 -- Let's wait a bit
	need_scan = true
end

function f:UNIT_QUEST_LOG_CHANGED(message, unit)
--err("UNIT_QUEST_LOG_CHANGED: message=%s, unit=%s", message, unit)
	if unit == "player" then
		elapsed = 0 -- Let's wait a bit
		need_scan = true
	end
end

function f:UNIT_SPELLCAST_SENT(event, unit, spell, rank, target)
	busy = true
end

function f:UNIT_SPELLCAST_FAILED(event, unit, spell, rank)
	busy = nil
end

function f:UNIT_SPELLCAST_STOP(event, unit, spell, rank)
	busy = nil
end

function f:UNIT_SPELLCAST_INTERRUPTED(event, unit, spell, rank)
	busy = nil
end

local lock_picking_spell = GetSpellInfo(1804)
function f:UNIT_SPELLCAST_SUCCEEDED(event, unit, spell, rank)
	busy = nil
	if unit == "player" and spell == lock_picking_spell then
		-- A something just got unlocked, it can be opened
		elapsed = -UPDATEPERIOD -- Let's wait a bit
		need_scan = true
		--Print("Something was unlocked")
	end
end

function f:UPDATE_CHAT_WINDOWS()
	need_chat_update = true
end

function f:PLAYER_LEVEL_UP()
	-- Clear the caches
	wipe(cache)
	wipe(do_not_cache)
	wipe(cache_clam)
end

local _PreviousRealZoneText = ""
function f:DetectZoneChange(event)
	-- Rescan the bags only if we are in the Garrison and the building changed
	-- local zoneID = GetCurrentMapAreaID()
	-- local current_zone =  GetRealZoneText()
	-- err("%s: zoneID = %s, current_zone = %s, _PreviousRealZoneText = %s", event, zoneID, current_zone, _PreviousRealZoneText)
	if GetCurrentMapAreaID() == 971 and _PreviousRealZoneText ~= GetRealZoneText() then
		_PreviousRealZoneText = GetRealZoneText()
		elapsed = 0 -- Let's wait a bit
		need_scan = true
	end
end
f.ZONE_CHANGED 				= f.DetectZoneChange
f.ZONE_CHANGED_INDOORS 		= f.DetectZoneChange
f.ZONE_CHANGED_NEW_AREA 	= f.DetectZoneChange

-----------------------------
--      Slash Handler      --
-----------------------------

_G.SLASH_SHUCKIT1 = "/shuckit"
_G.SlashCmdList.SHUCKIT = function(msg)
	Print(L["Use %s to open the next item"],"|cffffd700/click ShuckItFrame|r")
end

-----------------------------
--        LDB Stuff        --
-----------------------------

-- Sort function
local function sort_by_item_name (a,b)
	-- Sort the crate at the end
	if isSalvageItem[a] and not isSalvageItem[b] then
		return false
	elseif not isSalvageItem[a] and isSalvageItem[b] then
		return true
	end

	local name_a = GetItemInfo(a)
	local name_b = GetItemInfo(b)
	return name_a < name_b
end


local index = {}
function ShuckIt:DrawTooltip(anchor)
	self.tooltip_anchor = anchor or self.tooltip_anchor

	if not self.tooltip then
		self.tooltip = lqt:Acquire("ShuckItlqt", 1, "LEFT")
	end
	local tooltip = self.tooltip

	tooltip:Clear()
	tooltip:SmartAnchorTo(self.tooltip_anchor)
	tooltip:SetScale(1)

	local line, column = tooltip:AddHeader()
	tooltip:SetCell(line, 1, ("|cffffd700%s v%s (r%d)|r"):format(L["ShuckIt"], ShuckIt.version, ShuckIt.rev), "CENTER", 1)

	-- clean index
	for key in pairs(index) do index[key] = nil end

	-- Sort by item name
	for key in pairs(shuck_table) do tinsert(index,key) end
	sort(index, sort_by_item_name)

	if #index > 0 then
		line, column = tooltip:AddHeader()
		local item = L["item"]
		if shuck_count > 1 then item = L["items"] end
		tooltip:SetCell(line, 1, (L["|cffffd700%d %s to open|r"]):format(shuck_count,item), "CENTER", 1)
		tooltip:AddSeparator()
	end

	for _, link in ipairs (index) do
		local count = shuck_table[link]
		local name, _, _, _, _, _, _, _, _, texture = GetItemInfo(link)

		line, column = tooltip:AddLine()
		tooltip:SetCell(line, 1, (L["|T%s:16|t %s x %d"]):format(
												  texture,
												  name,
												  count
										 ), "LEFT")
		tooltip:SetCellScript(line, column, "OnMouseUp", function()
			if not InCombatLockdown() and not IsStealthed() and not busy and IsOpenable(shuck_table_bag[link], shuck_table_slot[link]) then
				UseContainerItem(shuck_table_bag[link], shuck_table_slot[link])
				return true
			elseif not InCombatLockdown() and not IsStealthed() and not busy and IsAClam(shuck_table_bag[link], shuck_table_slot[link]) then
				Print(L["Cannot open Clams by clicking on the LDB display. Use %s instead."], "|cffffd700/click ShuckItFrame|r")
			end
 		end)
	end

	tooltip:Show()

end

function dataobj:OnEnter(motion)
	ShuckIt:DrawTooltip(self)
	tooltip_elapsed = 0
end

function dataobj:OnLeave()
   --lqt:Release(ShuckIt.tooltip)
   --ShuckIt.tooltip = nil
end

----------------------------
--      Secure frame      --
----------------------------

do

	local frame = nil
	function ShuckIt:InitSecureFrame()
		--err("ShuckIt:InitSecureFrame")
		if InCombatLockdown() or frame then

			if not frame then

				frame = CreateFrame("frame","SkuckIt_InitSecureFrame")
				assert(frame, "Was not able to create frame")

				frame:Hide()

				frame:SetScript("OnUpdate", function(self, elap)
						if not InCombatLockdown then
							ShuckIt:InitSecureFrame()
							frame:Hide()
							frame = nil
						end
				end)
				frame:Show()
			end

			return
		end

		secureframe = CreateFrame("Button", "ShuckItFrame", UIParent, "SecureActionButtonTemplate")

		secureframe:SetScript("PreClick", function(self)
			if InCombatLockdown() or IsStealthed() or busy or
					(not IsOpenable(shuck_bag, shuck_slot) and not IsAClam(shuck_bag, shuck_slot)) then
				Print(L["Nothing to do or in combat"])
				return
			end
			self:SetAttribute( "type1", "macro" )
			self:SetAttribute( "macrotext1", "/use "..shuck_bag.." "..shuck_slot)
		end)


		secureframe:SetScript("PostClick", function(self)
			if InCombatLockdown() then return end
			self:SetAttribute( "type1", ATTRIBUTE_NOOP )
			self:SetAttribute( "macrotext1", nil)
		end)
		--err("Frame created: %s",secureframe)
	end

end